package listautenti_test

import (
	"context"
	"fmt"
	"io"
	"net"
	"net/http"
	"testing"

	"gotest.tools/assert"

	"github.com/valyala/fasthttp"
	"github.com/valyala/fasthttp/fasthttputil"

	"git.phc.dm.unipi.it/phc/website/libs/db"
	"git.phc.dm.unipi.it/phc/website/libs/sl"
	"git.phc.dm.unipi.it/phc/website/libs/util"
	"git.phc.dm.unipi.it/phc/website/server"

	"git.phc.dm.unipi.it/phc/website/server/config"
	"git.phc.dm.unipi.it/phc/website/server/database"
	"git.phc.dm.unipi.it/phc/website/server/listautenti"
	"git.phc.dm.unipi.it/phc/website/server/model"
)

func TestApiListaUtenti(t *testing.T) {

	memDB := &database.Memory{
		Users: []model.User{
			{
				Id:       "e39ad8d5-a087-4cb2-8fd7-5a6ca3f6a534",
				Username: "claire-doe",
				FullName: db.NewOption("Claire Doe"),
				Email:    db.NewOption("claire.doe@example.org"),
			},
			{
				Id:       "9b7109cd-95a1-41e9-a9f6-001a32c20ca1",
				Username: "john-smith",
				FullName: db.NewOption("John Smith"),
				Email:    db.NewOption("john.smith@example.org"),
			},
		},
	}

	l := sl.New()

	// Config
	sl.Provide(l, config.Slot, config.ExampleProductionConfig)

	// Database
	sl.Provide[database.Database](l, database.Slot, memDB)

	// Server
	sl.ProvideFunc(l, server.Slot, server.Configure)
	sl.ProvideHook(l, server.ApiRoutesHook,
		listautenti.MountApiRoutesHook,
	)

	// Initialize server instance
	srv, err := sl.Use(l, server.Slot)
	assert.NilError(t, err)

	//
	// Try doing the request
	//

	req, err := http.NewRequest("GET", "http://localhost:4000/api/lista-utenti", nil)
	assert.NilError(t, err)

	ln := fasthttputil.NewInmemoryListener()
	defer ln.Close()

	go func() {
		err := fasthttp.Serve(ln, srv.Router.Handler())
		if err != nil {
			panic(fmt.Errorf("failed to serve: %v", err))
		}
	}()

	client := http.Client{
		Transport: &http.Transport{
			DialContext: func(ctx context.Context, network, addr string) (net.Conn, error) {
				return ln.Dial()
			},
		},
	}

	res, err := client.Do(req)
	assert.NilError(t, err)

	body, err := io.ReadAll(res.Body)
	assert.NilError(t, err)

	assert.Equal(t, string(body), util.CompactIndentedLines(`
		[
			{
				"Id":"e39ad8d5-a087-4cb2-8fd7-5a6ca3f6a534",
				"Username":"claire-doe",
				"FullName":"Claire Doe",
				"Email":"claire.doe@example.org"
			},
			{
				"Id":"9b7109cd-95a1-41e9-a9f6-001a32c20ca1",
				"Username":"john-smith",
				"FullName":"John Smith",
				"Email":"john.smith@example.org"
			}
		]
	`))
}
